# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymaid/ext/monitor/monitor.proto
"""Generated protocol buffer code."""
from pymaid.rpc.pb import pymaid_pb2 as pymaid_dot_rpc_dot_pb_dot_pymaid__pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n pymaid/ext/monitor/monitor.proto\x12\x12pymaid.ext.monitor\x1a\x1apymaid/rpc/pb/pymaid.proto\"\x06\n\x04Pong2G\n\x0eMonitorService\x12\x35\n\x04Ping\x12\x13.pymaid.rpc.pb.Void\x1a\x18.pymaid.ext.monitor.PongB\x03\x90\x01\x01\x62\x06proto3')


_PONG = DESCRIPTOR.message_types_by_name['Pong']
Pong = _reflection.GeneratedProtocolMessageType('Pong', (_message.Message,), {
    'DESCRIPTOR': _PONG,
    '__module__': 'pymaid.ext.monitor.monitor_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.ext.monitor.Pong)
})
_sym_db.RegisterMessage(Pong)

_MONITORSERVICE = DESCRIPTOR.services_by_name['MonitorService']
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\220\001\001'
    _PONG._serialized_start = 84
    _PONG._serialized_end = 90
    _MONITORSERVICE._serialized_start = 92
    _MONITORSERVICE._serialized_end = 163
MonitorService = service_reflection.GeneratedServiceType('MonitorService', (_service.Service,), dict(
    DESCRIPTOR=_MONITORSERVICE,
    __module__='pymaid.ext.monitor.monitor_pb2'
))

MonitorService_Stub = service_reflection.GeneratedServiceStubType('MonitorService_Stub', (MonitorService,), dict(
    DESCRIPTOR=_MONITORSERVICE,
    __module__='pymaid.ext.monitor.monitor_pb2'
))


# @@protoc_insertion_point(module_scope)



import sys

from . import conf
from . import worker
from .parser import get_parser


__all__ = ['conf', 'worker']


def main():
    parser = get_parser()
    args = parser.parse_args()

    if args.debug:
        print(args)

    if not hasattr(args, 'entry'):
        print(parser.parse_args(sys.argv[1:] + ['-h']))
        exit(1)

    args.entry(args)



import json

from pymaid.conf import settings

from .parser import get_parser


def entry(args):
    data = settings.namespaces
    if args.list_ns:
        data = list(settings.namespaces)
    elif args.list_types:
        data = list(settings.transformer)
    elif args.ns:
        try:
            data = settings.namespaces[args.ns]
        except KeyError:
            print(
                f'ns `{args.ns}` not found, available namespaces: '
                f'{list(settings.namespaces)}'
            )
            exit(1)

    if args.format == 'json':
        print(json.dumps(data, ensure_ascii=False, indent=4))
    if args.format == 'py':
        import pprint
        pprint.pprint(data)


parser = get_parser().create_subparser('conf', help='show settings data')
parser.add_argument(
    '--list-ns',
    action='store_true',
    help='show the namespaces of settings',
)
parser.add_argument(
    '--ns',
    type=str,
    help='only show the settings of specified namespace',
)
parser.add_argument(
    '--list-types',
    action='store_true',
    help='show the available types for conf',
)
parser.add_argument(
    '--format',
    choices=('json', 'py'),
    default='json',
    help='specify the output format',
)
parser.set_defaults(entry=entry)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymaid/rpc/pb/pymaid.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1apymaid/rpc/pb/pymaid.proto\x12\rpymaid.rpc.pb\"\x9b\x03\n\x07\x43ontext\x12\x17\n\x0ftransmission_id\x18\x01 \x01(\r\x12\x36\n\x0bpacket_type\x18\x02 \x01(\x0e\x32!.pymaid.rpc.pb.Context.PacketType\x12\x37\n\x0cpacket_flags\x18\x03 \x01(\x0e\x32!.pymaid.rpc.pb.Context.PacketFlag\x12\x31\n\x08priority\x18\x04 \x01(\x0e\x32\x1f.pymaid.rpc.pb.Context.Priority\x12\x16\n\x0eservice_method\x18\x05 \x01(\t\x12\x14\n\x0cis_cancelled\x18\x06 \x01(\x08\x12\x11\n\tis_failed\x18\x07 \x01(\x08\"4\n\nPacketType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07REQUEST\x10\x01\x12\x0c\n\x08RESPONSE\x10\x02\"4\n\nPacketFlag\x12\x08\n\x04NULL\x10\x00\x12\x07\n\x03NEW\x10\x01\x12\n\n\x06\x43\x41NCEL\x10\x02\x12\x07\n\x03\x45ND\x10\x04\"&\n\x08Priority\x12\x07\n\x03LOW\x10\x00\x12\x07\n\x03MID\x10\x01\x12\x08\n\x04HIGH\x10\x02\"\x08\n\x06RpcAck\"\x06\n\x04Void\";\n\x0c\x45rrorMessage\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\tb\x06proto3')


_CONTEXT = DESCRIPTOR.message_types_by_name['Context']
_RPCACK = DESCRIPTOR.message_types_by_name['RpcAck']
_VOID = DESCRIPTOR.message_types_by_name['Void']
_ERRORMESSAGE = DESCRIPTOR.message_types_by_name['ErrorMessage']
_CONTEXT_PACKETTYPE = _CONTEXT.enum_types_by_name['PacketType']
_CONTEXT_PACKETFLAG = _CONTEXT.enum_types_by_name['PacketFlag']
_CONTEXT_PRIORITY = _CONTEXT.enum_types_by_name['Priority']
Context = _reflection.GeneratedProtocolMessageType('Context', (_message.Message,), {
    'DESCRIPTOR': _CONTEXT,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.Context)
})
_sym_db.RegisterMessage(Context)

RpcAck = _reflection.GeneratedProtocolMessageType('RpcAck', (_message.Message,), {
    'DESCRIPTOR': _RPCACK,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.RpcAck)
})
_sym_db.RegisterMessage(RpcAck)

Void = _reflection.GeneratedProtocolMessageType('Void', (_message.Message,), {
    'DESCRIPTOR': _VOID,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.Void)
})
_sym_db.RegisterMessage(Void)

ErrorMessage = _reflection.GeneratedProtocolMessageType('ErrorMessage', (_message.Message,), {
    'DESCRIPTOR': _ERRORMESSAGE,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.ErrorMessage)
})
_sym_db.RegisterMessage(ErrorMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CONTEXT._serialized_start = 46
    _CONTEXT._serialized_end = 457
    _CONTEXT_PACKETTYPE._serialized_start = 311
    _CONTEXT_PACKETTYPE._serialized_end = 363
    _CONTEXT_PACKETFLAG._serialized_start = 365
    _CONTEXT_PACKETFLAG._serialized_end = 417
    _CONTEXT_PRIORITY._serialized_start = 419
    _CONTEXT_PRIORITY._serialized_end = 457
    _RPCACK._serialized_start = 459
    _RPCACK._serialized_end = 467
    _VOID._serialized_start = 469
    _VOID._serialized_end = 475
    _ERRORMESSAGE._serialized_start = 477
    _ERRORMESSAGE._serialized_end = 536
# @@protoc_insertion_point(module_scope)



'''
0                   1                   2                   3
0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-------+-+-------------+-------------------------------+
|F|R|R|R| opcode|M| Payload len |    Extended payload length    |
|I|S|S|S|  (4)  |A|     (7)     |             (16/64)           |
|N|V|V|V|       |S|             |   (if payload len==126/127)   |
| |1|2|3|       |K|             |                               |
+-+-+-+-+-------+-+-------------+ - - - - - - - - - - - - - - - +
|     Extended payload length continued, if payload len == 127  |
+ - - - - - - - - - - - - - - - +-------------------------------+
|                               |Masking-key, if MASK set to 1  |
+-------------------------------+-------------------------------+
| Masking-key (continued)       |          Payload Data         |
+-------------------------------- - - - - - - - - - - - - - - - +
:                     Payload Data continued ...                :
+ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
|                     Payload Data continued ...                |
+---------------------------------------------------------------+
'''

import struct

from base64 import b64encode
from enum import IntEnum
from hashlib import sha1
from typing import List, Tuple

from multidict import CIMultiDict
from pymaid.net.protocol import Protocol
from pymaid.types import DataType

from .exceptions import ProtocolError

try:
    from .speedups import apply_mask
except ImportError:

    # https://www.willmcgugan.com/blog/tech/post/speeding-up-websockets-60x/

    _XOR_TABLE = [bytes(a ^ b for a in range(256)) for b in range(256)]

    def apply_mask(payload: DataType, mask: bytes) -> bytes:
        if len(mask) != 4:
            raise ValueError('mask must be 4 bytes')

        a, b, c, d = (_XOR_TABLE[n] for n in mask)
        data_bytes = bytearray(payload)
        data_bytes[::4] = data_bytes[::4].translate(a)
        data_bytes[1::4] = data_bytes[1::4].translate(b)
        data_bytes[2::4] = data_bytes[2::4].translate(c)
        data_bytes[3::4] = data_bytes[3::4].translate(d)
        return bytes(data_bytes)


__all__ = ['WSProtocol', 'Frame', 'apply_mask']

st = struct.Struct('!BB')
pack_header = st.pack
unpack_header = st.unpack

st_H = struct.Struct('!H')
pack_H = st_H.pack
unpack_H = st_H.unpack

st_Q = struct.Struct('!Q')
pack_Q = st_Q.pack
unpack_Q = st_Q.unpack


class CloseReason(IntEnum):
    '''RFC 6455, Section 7.4.1 - Defined Status Codes'''

    #: indicates a normal closure, meaning that the purpose for
    #: which the connection was established has been fulfilled.
    NORMAL_CLOSURE = 1000

    #: indicates that an endpoint is 'going away', such as a server
    #: going down or a browser having navigated away from a page.
    GOING_AWAY = 1001

    #: indicates that an endpoint is terminating the connection due
    #: to a protocol error.
    PROTOCOL_ERROR = 1002

    #: indicates that an endpoint is terminating the connection
    #: because it has received a type of data it cannot accept (e.g., an
    #: endpoint that understands only text data MAY send this if it
    #: receives a binary message).
    UNSUPPORTED_DATA = 1003

    #: Reserved.  The specific meaning might be defined in the future.
    # DON'T DEFINE THIS: RESERVED_1004 = 1004

    #: is a reserved value and MUST NOT be set as a status code in a
    #: Close control frame by an endpoint.  It is designated for use in
    #: applications expecting a status code to indicate that no status
    #: code was actually present.
    NO_STATUS_RCVD = 1005

    #: is a reserved value and MUST NOT be set as a status code in a
    #: Close control frame by an endpoint.  It is designated for use in
    #: applications expecting a status code to indicate that the
    #: connection was closed abnormally, e.g., without sending or
    #: receiving a Close control frame.
    ABNORMAL_CLOSURE = 1006

    #: indicates that an endpoint is terminating the connection
    #: because it has received data within a message that was not
    #: consistent with the type of the message (e.g., non-UTF-8 [RFC3629]
    #: data within a text message).
    INVALID_FRAME_PAYLOAD_DATA = 1007

    #: indicates that an endpoint is terminating the connection
    #: because it has received a message that violates its policy.  This
    #: is a generic status code that can be returned when there is no
    #: other more suitable status code (e.g., 1003 or 1009) or if there
    #: is a need to hide specific details about the policy.
    POLICY_VIOLATION = 1008

    #: indicates that an endpoint is terminating the connection
    #: because it has received a message that is too big for it to
    #: process.
    MESSAGE_TOO_BIG = 1009

    #: indicates that an endpoint (client) is terminating the
    #: connection because it has expected the server to negotiate one or
    #: more extension, but the server didn't return them in the response
    #: message of the WebSocket handshake.  The list of extensions that
    #: are needed SHOULD appear in the /reason/ part of the Close frame.
    #: Note that this status code is not used by the server, because it
    #: can fail the WebSocket handshake instead.
    MANDATORY_EXT = 1010

    #: indicates that a server is terminating the connection because
    #: it encountered an unexpected condition that prevented it from
    #: fulfilling the request.
    INTERNAL_ERROR = 1011

    #: is a reserved value and MUST NOT be set as a status code in a
    #: Close control frame by an endpoint.  It is designated for use in
    #: applications expecting a status code to indicate that the
    #: connection was closed due to a failure to perform a TLS handshake
    #: (e.g., the server certificate can't be verified).
    TLS_HANDSHAKE_FAILED = 1015


# RFC 6455, Section 7.4.1 - Defined Status Codes
LOCAL_ONLY_CLOSE_REASONS = {
    CloseReason.NO_STATUS_RCVD,
    CloseReason.ABNORMAL_CLOSURE,
    CloseReason.TLS_HANDSHAKE_FAILED,
}


# RFC 6455, Section 7.4.2 - Status Code Ranges
MIN_CLOSE_REASON = 1000
MIN_PROTOCOL_CLOSE_REASON = 1000
MAX_PROTOCOL_CLOSE_REASON = 2999
MIN_LIBRARY_CLOSE_REASON = 3000
MAX_LIBRARY_CLOSE_REASON = 3999
MIN_PRIVATE_CLOSE_REASON = 4000
MAX_PRIVATE_CLOSE_REASON = 4999
MAX_CLOSE_REASON = 4999


class WSProtocol(Protocol):

    HANDSHAKE_REQ = (
        b'GET %s HTTP/1.1\r\n'
        b'Host: %s\r\n'
        b'Upgrade: WebSocket\r\n'
        b'Connection: Upgrade\r\n'
        b'Sec-WebSocket-Key: %s\r\n'
        b'Sec-WebSocket-Version: %s\r\n'
    )
    HANDSHAKE_RESP = (
        b'HTTP/1.1 101 Switching Protocols\r\n'
        b'Upgrade: WebSocket\r\n'
        b'Connection: Upgrade\r\n'
        b'Sec-WebSocket-Accept: %s\r\n\r\n'
    )

    VERSION = b'13'
    SUPPORTED_VERSIONS = ('13', '8', '7')
    GUID = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

    MAX_HEADER_SIZE = 4096

    @classmethod
    def feed_data(cls, buf: memoryview) -> Tuple[int, List['Frame']]:
        frames = []
        buf = memoryview(buf)

        total_used = 0
        while 1:
            try:
                used_size, frame = Frame.decode(buf)
            except ProtocolError as ex:
                frames.append(ex)
                break
            if not frame:
                break
            frames.append(frame)
            total_used += used_size
            buf = buf[used_size:]

        return total_used, frames

    @classmethod
    def encode_frame(
        cls,
        opcode: int,
        data: DataType,
        mask: bytes = b''
    ) -> bytes:
        # no mater what opcode, message should be binary type
        if isinstance(data, str):
            data = data.encode('utf-8')
        return Frame.encode(opcode, data, mask)

    @classmethod
    def encode(cls, data: DataType, mask: bytes = b'') -> bytes:
        '''Encode a frame with data as its payload.

        :params data: data to send
        '''
        return cls.encode_frame(
            isinstance(data, str) and Frame.OPCODE_TEXT or Frame.OPCODE_BINARY,
            data,
            mask,
        )

    @classmethod
    def decode(cls, buf: memoryview) -> Tuple[int, 'Frame']:
        '''Decode a `Frame` from the buffer.

        :return: The consumed data length and decoded frame as a tuple.
        '''
        return Frame.decode(buf)

    @classmethod
    def build_request(cls, hostname, resource, key, **headers):
        basic = cls.HANDSHAKE_REQ % (resource, hostname, key, cls.VERSION)
        extra = b''.join(
            b'%s: %s\r\n' % (
                key if isinstance(key, bytes) else key.encode('utf-8'),
                value if isinstance(value, bytes) else value.encode('utf-8')
            )
            for key, value in headers.items()
        )
        return b'%s%s\r\n' % (basic, extra)

    @classmethod
    def build_response(cls, headers: CIMultiDict) -> bytes:
        if ('websocket' != headers.get('Upgrade', '').lower()
                or 'upgrade' != headers.get('Connection', '').lower()):
            raise ProtocolError(
                f'invalid websocket handshake header: {headers}'
            )

        version = headers.get('Sec-WebSocket-Version')
        if version not in cls.SUPPORTED_VERSIONS:
            raise ProtocolError(f'unsupported version={version}')

        sec_key = headers.get('Sec-WebSocket-Key', '')
        if not sec_key:
            raise ProtocolError('missing Sec-WebSocket-Key')

        # TODO: check origin; check host; check headers; check ext

        resp_key = b64encode(sha1(sec_key.encode('utf-8') + cls.GUID).digest())
        return cls.HANDSHAKE_RESP % resp_key

    @classmethod
    def validate_upgrade(cls, headers: CIMultiDict, upgrade_key: bytes):
        if ('websocket' != headers.get('Upgrade', '').lower()
                or 'upgrade' != headers.get('Connection', '').lower()):
            raise ProtocolError(
                f'invalid websocket handshake header: {headers}'
            )

        accept = headers.get('Sec-WebSocket-Accept', '').lower()
        if isinstance(accept, str):
            accept = accept.encode('utf-8')
        value = upgrade_key + cls.GUID
        if b64encode(sha1(value).digest()).strip().lower() != accept:
            raise ProtocolError('invalid accept value')


class Frame:

    __slots__ = (
        'fin', 'opcode', 'flags', 'length', 'mask', 'payload', 'close_reason',
    )

    FIN_MASK = 0x80
    OPCODE_MASK = 0x0f
    MASK_MASK = 0x80
    LENGTH_MASK = 0x7f

    RSV0_MASK = 0x40
    RSV1_MASK = 0x20
    RSV2_MASK = 0x10

    # bitwise mask that will determine the reserved bits for a frame header
    RSV_MASK = RSV0_MASK | RSV1_MASK | RSV2_MASK

    OPCODE_CONTINUATION = 0x00
    OPCODE_TEXT = 0x01
    OPCODE_BINARY = 0x02
    OPCODE_CLOSE = 0x08
    OPCODE_PING = 0x09
    OPCODE_PONG = 0x0a

    def __init__(
        self,
        opcode: int,
        payload: DataType,
        mask: bytes = b'',
        fin: bool = True,
        flags: bytes = b'',
        length: int = 0,
    ):
        if mask:
            payload = apply_mask(payload, mask)
        self.opcode = opcode
        self.payload = payload
        self.mask = mask
        self.fin = fin
        self.flags = flags
        self.length = length or len(payload)
        self.close_reason = None

        self.prepare()

    def prepare(self):
        if self.opcode == self.OPCODE_CLOSE:
            payload = self.payload
            length = self.length
            if not length:
                self.close_reason = CloseReason.NO_STATUS_RCVD
            elif length == 1:
                raise ProtocolError(f'Invalid close frame: {self} {payload}')
            else:
                code = unpack_H(payload[:2])[0]
                if code < MIN_CLOSE_REASON or code > MAX_CLOSE_REASON:
                    raise ProtocolError('invalid close code range')
                try:
                    code = CloseReason(code)
                except ValueError:
                    pass
                if code in LOCAL_ONLY_CLOSE_REASONS:
                    raise ProtocolError('remote CLOSE with local-only reason')
                if (not isinstance(code, CloseReason)
                        and code <= MAX_PROTOCOL_CLOSE_REASON):
                    raise ProtocolError('CLOSE with unknown reserved code')
                try:
                    reason = payload[2:].decode('utf-8')
                except UnicodeDecodeError:
                    raise ProtocolError(
                        'close reason is not valid UTF-8',
                        CloseReason.INVALID_FRAME_PAYLOAD_DATA,
                    )
                if isinstance(code, CloseReason):
                    code.reason = reason
                else:
                    code = (code, reason)
                self.close_reason = code

    def __repr__(self):
        return (
            f'<Frame fin={self.fin} opcode={self.opcode} flags={self.flags} '
            f'length={self.length} mask={self.mask.hex()}>'
        )

    @classmethod
    def encode(
        cls,
        opcode: int,
        data: bytes,
        mask: bytes = b'',
        fin: bool = True,
        flags: bytes = b'',
    ) -> bytes:
        '''Encodes a WebSocket frame.

        :param opcode: The opcode of the payload, see `OPCODE_*`
        :param data: The payload data.
        :param mask: The payload of mask key.
        :param fin: Whether this is the final frame for this opcode.
        :param flags: The RSV* flags.
        :returns: A bytestring encoded frame.
        '''
        first_byte = opcode
        second_byte = 0
        extra = b''

        if fin:
            first_byte |= cls.FIN_MASK
        if flags:
            first_byte |= (flags & cls.RSV_MASK)

        # now deal with length complexities
        length = len(data)
        if length < 126:
            second_byte |= length
        elif length <= 0xffff:
            second_byte |= 126
            extra = pack_H(length)
        elif length <= 0xffffffffffffffff:
            second_byte |= 127
            extra = pack_Q(length)
        else:
            raise ProtocolError(length, CloseReason.MESSAGE_TOO_BIG)

        if mask:
            second_byte |= cls.MASK_MASK
            extra += mask
            data = apply_mask(data, mask)
        return pack_header(first_byte, second_byte) + extra + data

    @classmethod
    def decode(cls, buf: memoryview) -> Tuple[int, 'Frame']:
        '''Decode a WebSocket frame.

        :param data: `DataType`
        :returns: consumed data length and decoded `Frame` instance.
        '''
        if len(buf) < 2:
            return 0, None

        header = buf[:2]
        first_byte, second_byte = unpack_header(header)
        fin = first_byte & cls.FIN_MASK == cls.FIN_MASK
        opcode = first_byte & cls.OPCODE_MASK
        length = second_byte & cls.LENGTH_MASK
        flags = first_byte & cls.RSV_MASK
        has_mask = second_byte & cls.MASK_MASK == cls.MASK_MASK

        if flags:
            raise ProtocolError(f'invalid flags: {flags}')

        if opcode > 0x07:
            if not fin:
                raise ProtocolError('Received fragmented control frame')
            # Control frames MUST have a payload length of 125 bytes or less
            if length > 125:
                raise ProtocolError(
                    f'Control frame cannot be larger than 125 bytes: {length}'
                )

        used_size = 2
        if length == 126:
            # extended payload length: 16 bits
            epl = buf[used_size: used_size + 2]
            if len(epl) < 2:
                return 0, None
            length = unpack_H(epl)[0]
            used_size += 2
        elif length == 127:
            # extended payload length: 64 bits
            epl = buf[used_size: used_size + 8]
            if len(epl) < 8:
                return 0, None
            length = unpack_Q(epl)[0]
            used_size += 8

        mask = b''
        if has_mask:
            mask = buf[used_size: used_size + 4].tobytes()
            if len(mask) < 4:
                return 0, None
            used_size += 4

        payload = buf[used_size: used_size + length]
        if len(payload) < length:
            return 0, None
        used_size += length

        return used_size, cls(
            opcode=opcode,
            payload=payload,
            mask=mask,
            fin=fin,
            flags=flags,
            length=length,
        )



import atexit
import fcntl
import io
import json
import multiprocessing
import os
import sys
import tempfile
import time

from functools import partial
from pathlib import Path
from psutil import pid_exists
from typing import Callable

from pymaid.utils.logger import get_logger

logger = get_logger('daemon')


def redirect_streams(stdin: str, stdout: str, stderr: str):
    '''Redirect stdin/stdout/stderr streams to the passed files'''
    sys.stdout.flush()
    sys.stderr.flush()

    si = Path(stdin)
    if not si.exists():
        si.parent.mkdir(0o700, parents=True, exist_ok=True)
        si.touch()
    si = si.open('r')

    so = Path(stdout)
    if not so.exists():
        so.parent.mkdir(0o700, parents=True, exist_ok=True)
        so.touch()
    so = so.open('a+')

    se = Path(stderr)
    if not se.exists():
        se.parent.mkdir(0o700, parents=True, exist_ok=True)
        se.touch()
    se = se.open('ab+', 0)

    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


def cleanup(
    name: str,
    pid_path: Path,
    pid_file: io.FileIO,
    stdin: str,
    stdout: str,
    stderr: str,
    clean_stream_files: bool = False,
):
    logger.info(
        f'[daemon] cleanup [name|{name}][pid_path|{pid_path}]'
        f'[stdin|{stdin}][stdout|{stdout}][stderr|{stderr}]'
    )
    fcntl.lockf(pid_file.fileno(), fcntl.LOCK_UN)
    pid_path.unlink(True)
    unrecord_worker([name])
    if clean_stream_files:
        for name in (stdin, stdout, stderr):
            Path(name).unlink(True)


def record_worker(workers):
    account_book = Path(f'{tempfile.gettempdir()}/workers')
    if not account_book.exists():
        account_book.touch()
    records = json.loads(account_book.read_text() or '{}')
    records.update(workers)
    account_book.write_text(json.dumps(records))


def unrecord_worker(workers):
    account_book = Path(f'{tempfile.gettempdir()}/workers')
    if not account_book.exists():
        account_book.touch()
    records = json.loads(account_book.read_text() or '{}')
    for key in workers:
        records.pop(key, None)
    account_book.write_text(json.dumps(records))


def list_worker():
    account_book = Path(f'{tempfile.gettempdir()}/workers')
    if not account_book.exists():
        account_book.touch()
    return json.loads(account_book.read_text() or '{}')


def _daemonize(
    name: str,
    index: int,
    cwd: str,
    queue,
    donot_exist: bool = False,
):
    '''
    do the UNIX double-fork magic, see Stevens' "Advanced
    Programming in the UNIX Environment" for details (ISBN 0201563177)
    http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16

    :params donot_exist: return instead of exit
    :returns: True if donot_exist
    '''
    try:
        pid = os.fork()
        if pid > 0:
            if donot_exist:
                return False
            # exit first parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write('fork #1 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    # decouple from parent environment
    os.chdir(cwd)
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write('fork #2 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    prefix = f'{cwd}/{name}-{index}'

    stdin = f'{prefix}-stdin'
    stdout = f'{prefix}-stdout'
    stderr = f'{prefix}-stderr'
    try:
        pid = os.getpid()
        # write pidfile; add exclusive lock
        pid_path = Path(f'{prefix}-pid')
        if not pid_path.exists():
            pid_path.parent.mkdir(0o700, parents=True, exist_ok=True)
            pid_path.touch()
        pid_path.write_text('%s' % pid)
        pid_file = pid_path.open()
        fcntl.flock(pid_file.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)

        # output to pymaid console script std
        # logger.info(
        #     f'[pymaid|daemonize] [pid|{pid}][cwd|{cwd}]'
        #     f'[stdin|{stdin}][stdout|{stdout}][stderr|{stderr}]'
        # )

        # redirect standard file descriptors
        redirect_streams(stdin, stdout, stderr)

        atexit.register(
            partial(
                cleanup, name, pid_path, pid_file, stdin, stdout, stderr, False
            )
        )
        ok = True
        message = 'ok'
    except Exception as ex:
        ok = False
        message = str(ex)

    queue.put(
        {
            f'{name}-{index}': {
                'index': index, 'pid': pid, 'ok': ok, 'message': message,
            }
        }
    )
    del queue

    return ok


def daemonize(entry: Callable, name: str, count: int = 1):
    assert callable(entry), entry
    tmp_dir = Path(f'{tempfile.gettempdir()}/{name}/')
    if not tmp_dir.exists():
        tmp_dir.mkdir(0o700, parents=True)

    queue = multiprocessing.Queue()
    for idx in range(count):
        key = f'{name}-{idx}'
        if _daemonize(
            name,
            idx,
            tmp_dir,
            queue,
            donot_exist=True,
        ):
            entry()
            # exit forked process
            exit(0)

    logger.info('[pymaid|daemon] getting worker processes info')
    workers = {}
    for idx in range(count):
        try:
            info = queue.get(timeout=1)
        except multiprocessing.queues.Empty:
            logger.info(
                f'[pymaid|daemon][count|{idx + 1}] get worker info timeout'
            )
        else:
            workers.update(info)
    logger.info(f'[pymaid|daemon][worker|{workers}]')
    del queue

    logger.info('[pymaid|daemon] checking worker processes status')

    succeeded = {}
    failed = {}
    while len(succeeded) + len(failed) != count:
        time.sleep(0.2)
        for idx in range(count):
            key = f'{name}-{idx}'
            if key in succeeded or key in failed:
                continue
            info = workers.get(key)
            if not info:
                failed[key] = 'not started'
                continue

            if info['ok']:
                if pid_exists(info['pid']):
                    succeeded[key] = info
                else:
                    try:
                        with open(f'{tmp_dir}/{key}-stderr', 'rb') as fp:
                            fp.seek(-256, 2)
                            err = '...' + fp.read().decode('utf-8')
                    except FileNotFoundError:
                        err = 'err not found'
                    failed[key] = err
            else:
                failed[key] = info['message']
    logger.info(
        f'[pymaid|daemon] workers [succeeded|{succeeded}][failed|{failed}]'
    )
    record_worker(succeeded)



import ssl as _ssl
from typing import Callable, List, Optional, Type, Union

from pymaid.ext.handler import SerialHandler
from pymaid.ext.middleware import MiddlewareManager
from pymaid.net import dial_stream as raw_dial_stream
from pymaid.net.protocol import ProtocolType
from pymaid.net.stream import Stream
from pymaid.rpc.connection import Connection, ConnectionType
from pymaid.types import HandlerType

from . import context
from . import protocol
from . import router

__all__ = ('connection', 'context', 'protocol')


async def dial_stream(
    address: str,
    *,
    transport_class: ConnectionType = Stream | Connection,
    ssl_context: Union[None, bool, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    on_open: Optional[List[Callable]] = None,
    on_close: Optional[List[Callable]] = None,
    # below are connection kwargs
    protocol_class: protocol.Protocol = protocol.Protocol,
    handler_class: HandlerType = SerialHandler,
    router_class: Type[router.Router] = router.PBRouter,
    context_manager_class: context.ContextManager = context.ContextManager,
    middleware_manager: Optional[MiddlewareManager] = None,
    **kwargs,
):
    return await raw_dial_stream(
        address,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        protocol=protocol_class(),
        handler=handler_class(),
        router=router_class(),
        context_manager=context_manager_class(initiative=True),
        **kwargs,
    )


async def serve_stream(
    address: str,
    *,
    transport_class: ConnectionType = Stream | Connection,
    protocol_class: ProtocolType = protocol.Protocol,
    handler_class: HandlerType = SerialHandler,
    router_class: Type[router.Router] = router.PBRouter,
    context_manager_class: context.ContextManager = context.ContextManager,
    **kwargs,
):
    from pymaid.rpc import serve_stream as raw_serve_stream
    return await raw_serve_stream(
        address=address,
        transport_class=transport_class,
        protocol_class=protocol_class,
        handler_class=handler_class,
        router_class=router_class,
        context_manager_class=context_manager_class,
        **kwargs,
    )


def implall(service):
    service_name = service.DESCRIPTOR.name
    missing = []
    for base in service.__bases__:
        for method in base.DESCRIPTOR.methods:
            method_name = method.name
            base_method = getattr(base, method_name)
            impl_method = getattr(service, method_name, base_method)
            if base_method == impl_method:
                missing.append(f'{service_name}.{method_name}')
    if missing:
        raise RuntimeError(f'{missing} are not implemented')
    return service



from pymaid.rpc.error import RPCError


PBError = RPCError.create_manager('PBError')
PBError.add_error('RPCNotFound', 'rpc not found')
PBError.add_error('InvalidTransmissionID', 'transmission_id value is invalid')
PBError.add_error('InvalidPacketType', 'cannot handle unknown packet')



import importlib
import json
import re
import sys
import tempfile

from pathlib import Path

from psutil import Process, NoSuchProcess

from pymaid.conf import settings
from pymaid.core import run, run_in_processpool, gather
from pymaid.core import CancelledError, ProcessPoolExecutor
from pymaid.utils.daemon import daemonize, list_worker
from pymaid.utils.logger import get_logger

from .parser import get_parser

path_to_mod_regex = re.compile(r'(.*)\.py[co]?$')

logger = get_logger('pymaid')


def entry_run(args):
    if args.main.count(':') != 1:
        print('main entry need to be formatted as `package.module:main_entry`')
        exit(1)
    mod, main = args.main.split(':')
    mod = mod.replace('/', '.')
    if path_to_mod_regex.match(mod):
        mod = path_to_mod_regex.match(mod).group(1)
    mod = importlib.import_module(mod)
    if args.pass_through:
        assert len(args.pass_through) == 1, args.pass_through
        sys.argv = [args.main] + args.pass_through[0]
    if args.parallel < 1:
        print(f'-p/--parallel should be positive, got {args.parallel}')
        exit(1)
    if args.parallel > 1:
        # enabled REUSE_PORT for parallel workers
        settings.pymaid.update({'REUSE_PORT': True})

    main_not_call = getattr(mod, main)
    if args.daemon:
        daemonize(
            lambda: run(
                main_not_call(*(args.args or ()), **(args.kwargs or {}))
            ),
            name=args.name or args.main,
            count=args.parallel,
        )
    elif args.parallel == 1:
        run(main_not_call(*args.args, **args.kwargs))
    else:
        async def wrapper():
            executor = ProcessPoolExecutor(args.parallel)
            results = [
                run_in_processpool(
                    run,
                    args=(main_not_call,),
                    kwargs={'args': args.args, 'kwargs': args.kwargs},
                    executor=executor,
                )
                for _ in range(args.parallel)
            ]
            try:
                await gather(*results, return_exceptions=False)
            except (SystemExit, KeyboardInterrupt, CancelledError):
                pass
        run(wrapper())


def stop_worker(pid):
    try:
        p = Process(int(pid))
    except NoSuchProcess:
        logger.warning(f'process {pid} not found')
        return False
    p.terminate()
    return True


def split_index_range(index):
    if index == 'all':
        for value in list_worker().values():
            yield value['index']
        return

    for idx in index.split(','):
        if '-' in idx:
            start, end = [int(i) for i in idx.split('-')]
            yield from range(start, end + 1)
        else:
            yield int(idx)


def entry_stop(args):
    if not args.name:
        print('manage need worker name, missing `--name` ?')
        exit(1)
    name = args.name
    tmp_dir = Path(f'{tempfile.gettempdir()}/{name}/')
    for index in split_index_range(args.index):
        pidfile = Path(f'{tmp_dir}/{name}-{index}-pid')
        if not pidfile.exists():
            logger.warning(f'worker {name}-{index} not found')
            continue
        pid = int(pidfile.read_text())
        ret = stop_worker(pid)
        logger.info(f'worker {name}-{index}|{pid} stopped: {ret}')


def entry_ls(args):
    print(json.dumps(list_worker(), indent=4))


worker_parser = get_parser().create_subparser('worker', help='worker')

parser_run = worker_parser.create_subparser('run', help='run the entry')
parser_run.add_argument(
    'main',
    type=str,
    metavar='{package.module:main | path/to/script.py:main}',
    help='main entry',
)
parser_run.add_argument(
    '--name',
    type=str,
    # default=''.join(random.sample(string.ascii_letters, 6)),
    help='specify worker name, if not set, args.main will be used',
)
parser_run.add_argument(
    '-d',
    '--daemon',
    action='store_true',
    default=False,
    help='run entry in daemon mode',
)
parser_run.add_argument(
    '-p',
    '--parallel',
    type=int,
    default=1,
    help='run entry parallelly',
)
parser_run.add_argument(
    '--args',
    type=json.loads,
    default=[],
    help='json formatted args passed to main entry',
)
parser_run.add_argument(
    '--kwargs',
    type=json.loads,
    default={},
    help='json formatted kwargs passed to main entry'
)
parser_run.add_argument(
    'pass_through',
    action='append',
    nargs='*',
    help=(
        'args directly pass through, '
        'if args startswith -/--, separate with `--`; '
        'e.g. "pymaid run a:main -- 1 2 --main_opt 3"'
    ),
)
parser_run.set_defaults(entry=entry_run)

parser_ls = worker_parser.create_subparser('ls', help='list the workers')
parser_ls.set_defaults(entry=entry_ls)

parser_stop = worker_parser.create_subparser(
    'stop', help='stop the worker',
)
parser_stop.add_argument('name', metavar='worker_name', help='worker name')
parser_stop.add_argument(
    'index',
    metavar='worker_index',
    nargs='?',
    default='all',
    help='e.g.: 1,2,3-5',
)
parser_stop.set_defaults(entry=entry_stop)



from socket import error as socket_error
from struct import Struct

st_H = Struct('!H')
pack_H = st_H.pack


class WebSocketError(socket_error):
    '''Base class for all websocket errors.'''


class ProtocolError(WebSocketError):
    '''Raised if an error occurs when de/encoding the websocket protocol.'''

    def __init__(self, reason='', code=1002):
        self.code = code
        self.reason = reason
        self.payload = pack_H(code) + reason.encode('utf-8')
